#include "EDLib.h"
#include <iostream>

extern "C" {
#include "iio.h"
}

using namespace std;

/*----------------------------------------------------------------------------*/
/*----------------------------- Write SVG File -------------------------------*/
/*----------------------------------------------------------------------------*/
/** Write line segments into a SVG file.
    If the name is "-" the file is written to standard output.
*/
static void write_svg(vector<LS> segs, char * filename, int xsize, int ysize, double width )
{
  FILE * svg;
  int i;

  /* open file */
  if( strcmp(filename,"-") == 0 ) svg = stdout;
  else svg = fopen(filename,"w");

  /* write SVG header */
  fprintf(svg,"<?xml version=\"1.0\" standalone=\"no\"?>\n");
  fprintf(svg,"<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"\n");
  fprintf(svg," \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n");
  fprintf(svg,"<svg width=\"%dpx\" height=\"%dpx\" ",xsize,ysize);
  fprintf(svg,"version=\"1.1\"\n xmlns=\"http://www.w3.org/2000/svg\" ");
  fprintf(svg,"xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n");

  /* write line segments */
  for(i=0;i<segs.size();i++)
    {
      fprintf(svg,"<line x1=\"%f\" y1=\"%f\" x2=\"%f\" y2=\"%f\" ",
              segs[i].start.x,segs[i].start.y,
              segs[i].end.x,segs[i].end.y);
      fprintf(svg,"stroke-width=\"%f\" stroke=\"black\" />\n", width);
    }

  /* close SVG file */
  fprintf(svg,"</svg>\n");
  fclose(svg);
}

static void write_eps( vector<LS> segs, char * filename, int xsize, int ysize, double width )
{
  FILE * eps;
  int i;


  /* open file */
  if( strcmp(filename,"-") == 0 ) eps = stdout;
  else eps = fopen(filename,"w");

  /* write EPS header */
  fprintf(eps,"%%!PS-Adobe-3.0 EPSF-3.0\n");
  fprintf(eps,"%%%%BoundingBox: 0 0 %d %d\n",xsize,ysize);
  fprintf(eps,"%%%%Creator: LSD, Line Segment Detector\n");
  fprintf(eps,"%%%%Title: (%s)\n",filename);
  fprintf(eps,"%%%%EndComments\n");

  /* write line segments */
  for(i=0;i<segs.size();i++)
    {
      fprintf( eps,"newpath %f %f moveto %f %f lineto %f setlinewidth stroke\n",
              segs[i].start.x,ysize - segs[i].start.y,
              segs[i].end.x,ysize - segs[i].end.y, width);
    }

  /* close EPS file */
  fprintf(eps,"showpage\n");
  fprintf(eps,"%%%%EOF\n");
  fclose(eps);
}

int main(int argc, char **argv)
{	
    if(argc < 3)
    {
        printf("Not enough parameters. It requires the input and output paths.");
        return 1;
    }

    int w, h, c;
    float* img = iio_read_image_float_vec(argv[1], &w, &h, &c);

    // TODO FUTURE WORK: change the order of the channels so that c is at the end when dealing with colors images

    // First test
    uchar* testImg = new uchar[w*h];
    for(int i = 0; i < w*h; ++i) 
    {
        float temp = 0.;
        for(int k = 0; k < c; ++k)
            temp += img[k + i*c];
        testImg[i] = (uchar) (temp/(float)c);
    }

	//Detection of line segments from the same image
	EDLines testEDLines = EDLines(testImg, w, h);
	//Mat lineImg = testEDLines.getLineImage();	//draws on an empty image

	//Acquiring line information, i.e. start & end points
	vector<LS> lines = testEDLines.getLines();
	int noLines = testEDLines.getLinesNo();
	std::cout << "Number of line segments: " << noLines << std::endl;

    write_svg(lines, argv[2], w, h, 1.5);
    write_eps(lines, argv[3], w, h, 1.5);

    FILE* output = fopen(argv[4],"w");
    for(int i=0;i<lines.size();i++)
        fprintf(output,"%f %f %f %f\n", lines[i].start.x, lines[i].start.y, lines[i].end.x, lines[i].end.y);
    fclose(output);

	return 0;
}



